from typing import Optional

import numpy as np


class SphereProperties(object):

    def __init__(self):
        self.center: Optional[np.ndarray] = None
        self.radius: Optional[float] = None

    def setCenter(self, center: np.ndarray) -> None:
        self.center = center

    def setRadius(self, radius: float) -> None:
        self.radius = radius

    def getCenter(self) -> np.ndarray:
        return self.center

    def getRadius(self) -> float:
        return self.radius



import math
from typing import TYPE_CHECKING, Union, List, Literal

import numpy as np

if TYPE_CHECKING:
    from geogeometry.geometry.vector.Vector import Vector


point_type = Union[List, np.ndarray]


class Angles(object):

    @staticmethod
    def calculateAngleBetweenVectorAndAxis(v: 'Vector', axis_id: Literal['x', 'y', 'z']) -> float:
        axes_vectors = {
            'x': [1, 0, 0],
            'y': [0, 1, 0],
            'z': [0, 0, 1]
        }
        return Angles.calculateAngleFromThreePoints(axes_vectors[axis_id], [0., 0., 0.], v.getTip())

    @staticmethod
    def calculateAngleBetweenPositionVectors(v1: 'Vector', v2: 'Vector') -> float:
        return Angles.calculateAngleFromThreePoints(v1.getTip(), [0., 0., 0.], v2.getTip())

    @staticmethod
    def calculateAngleFromThreePoints(n0: point_type, n1: point_type, n2: point_type) -> float:
        """
        Returns the angle of 2 3D segments sharing n1 node.
        Note: arccos domain: [-1,1]
        :returns angle in degrees
        """
        n0, n1, n2 = np.array([n0, n1, n2])
        v0 = n2 - n1
        v1 = n0 - n1

        if np.linalg.norm(v0) == 0. or np.linalg.norm(v1) == 0.:
            raise ValueError("Repeated points in angle calculation.")

        a = np.dot(v0, v1) / (np.linalg.norm(v0) * np.linalg.norm(v1))

        if a < -1.:
            a = -1.
        elif a > 1:
            a = 1

        arc = np.arccos(a)
        ang = math.degrees(arc)
        if np.isnan(ang):
            ang = 180.

        return ang



from typing import Optional, Literal, TYPE_CHECKING, Union, List

import numpy as np

if TYPE_CHECKING:
    from geogeometry.geometry.shared.explicit_geometry.ExplicitGeometryMetrics import ExplicitGeometryMetrics


class ExplicitGeometryProperties(object):

    def __init__(self):
        super().__init__()

        self.nodes: Optional[np.ndarray] = None

        self.explicit_metrics: Optional['ExplicitGeometryMetrics'] = None

    def setNodes(self, nodes: Union[List, np.ndarray]) -> None:
        if isinstance(nodes, list):
            nodes = np.array(nodes)
        self.nodes = nodes
        self.explicit_metrics.calculateMetrics()

    def resetNodes(self) -> None:
        """
        Without reseting metrics.
        Used when transforming into external library objects (e.g. trimesh).
        """
        self.nodes = None

    def getNodes(self) -> Optional[np.ndarray]:
        return self.nodes

    def getDimensions(self) -> Optional[Literal[2, 3]]:
        return self.explicit_metrics.getDimensions()

    def getLimits(self) -> Optional[np.ndarray]:
        return self.explicit_metrics.getLimits()

    def getCenter(self) -> Optional[np.ndarray]:
        return self.explicit_metrics.getCenter()



from typing import TYPE_CHECKING, Optional

import numpy as np

from geogeometry.geometry.shared.shared_components.BaseMetrics import BaseMetrics

if TYPE_CHECKING:
    from geogeometry.geometry.model.GeometryModel import GeometryModel


class GeometryModelMetrics(BaseMetrics):

    def __init__(self, model: 'GeometryModel'):
        self.model: 'GeometryModel' = model

        self.limits: Optional[np.ndarray] = None
        self.center: Optional[np.ndarray] = None

    def getLimits(self) -> Optional[np.ndarray]:
        return self.limits

    def getCenter(self) -> Optional[np.ndarray]:
        return self.center

    def calculateMetrics(self):
        self.calculateLimits()
        self.calculateCenter()

    def calculateLimits(self):
        all_limits = [lim for t in self.model.getTriangulations() for lim in t.getLimits()]
        all_limits += [lim for p in self.model.getPolylines() for lim in p.getLimits()]

        _min, _max = np.min(all_limits, axis=0), np.max(all_limits, axis=0)
        self.limits = np.array([_min, _max])

    def calculateCenter(self) -> None:
        # all_centers = [t.getCentroid() for t in self.model.getTriangulations()]
        # all_centers += [p.getCentroid() for p in self.model.getPolylines()]
        self.center = np.average(self.limits, axis=0)



from typing import Optional, TYPE_CHECKING, Literal, Union, List

import numpy as np

if TYPE_CHECKING:
    from geogeometry.geometry.vector.Vector import Vector
    from geogeometry.geometry.vector.components.VectorMetrics import VectorMetrics


class VectorProperties(object):

    def __init__(self):

        self.origin: Optional[np.ndarray] = None
        self.tip: Optional[np.ndarray] = None

        self.metrics: Optional['VectorMetrics'] = None

        # self.nodes: np.array = np.full((2, 3), np.nan)

        # self.length: Optional[float] = None

        # self.is_unitary: Optional[bool] = None
        # self.is_position_vector: Optional[bool] = None  # origin=(0,0,0)
        #
        # self.normalized_vector: Optional['Vector'] = None

        # self.dip: Optional[float] = None
        # self.dipdir: Optional[float] = None

    def setOrigin(self, origin: Union[List, np.ndarray]) -> None:
        if isinstance(origin, list):
            origin = np.array(origin)
        self.origin = origin
        # self.nodes[0] = origin

    def setTip(self, tip: Union[List, np.ndarray]) -> None:
        if isinstance(tip, list):
            tip = np.array(tip)
        self.tip = tip
        # self.nodes[1] = tip

    # def setNodes(self, nodes: Union[List, np.ndarray]) -> None:
    #     if len(nodes) != 2:
    #         raise ValueError("Trying to define an explicit vector without 2 nodes.")
    #
    #     self.setOrigin(origin=nodes[0])
    #     self.setTip(tip=nodes[1])

    # def setLength(self, length: float) -> None:
    #     self.length = length
    #
    # def setIsUnitary(self, is_unitary: bool) -> None:
    #     self.is_unitary = is_unitary
    #
    # def setIsPositionVector(self, is_position_vector: bool) -> None:
    #     self.is_position_vector = is_position_vector
    #
    # def setNormalizedVector(self, normalized_vector: 'Vector') -> None:
    #     self.normalized_vector = normalized_vector
    #
    # def setDip(self, dip: float) -> None:
    #     self.dip = dip
    #
    # def setDipdir(self, dipdir: float) -> None:
    #     self.dipdir = dipdir

    def getOrigin(self) -> np.ndarray:
        return self.origin

    def getTip(self) -> np.ndarray:
        return self.tip

    # def getNodes(self) -> np.ndarray:
    #     return self.nodes

    def getLength(self) -> float:
        return self.metrics.getLength()

    def isUnitary(self) -> bool:
        return self.metrics.isUnitary()

    def isPositionVector(self) -> bool:
        return self.metrics.isPositionVector()

    # def getNormalizedVector(self) -> 'Vector':
    #     return self.metrics.getNormalizedVector()

    def getUnitVector(self) -> np.ndarray:
        return self.metrics.getUnitVector()

    def getDip(self) -> float:
        return self.metrics.getDip()

    def getAzimuth(self) -> float:
        return self.metrics.getAzimuth()



from typing import TYPE_CHECKING

import pyvista as pv

from geogeometry.geometry.shared.shared_components.BasePlotter import BasePlotter

if TYPE_CHECKING:
    import matplotlib.pyplot as plt
    from geogeometry.geometry.model.GeometryModelsCollection import GeometryModelsCollection


class GeometryModelsCollectionPlotter(BasePlotter):

    def __init__(self, geometrymodels_collection: 'GeometryModelsCollection'):
        super().__init__(element=geometrymodels_collection)
        self.geometrymodels_collection: 'GeometryModelsCollection' = geometrymodels_collection

    def addToMatplotlibAx(self, ax: 'plt.Axes') -> None: ...

    def addToPyvistaPlotter(self, plotter: 'pv.Plotter', **kwargs) -> None:
        for m in self.geometrymodels_collection:
            m.addToPyvistaPlotter(plotter=plotter, **kwargs)

    def addToParaviewPlotter(self, plotter, **kwargs) -> None: ...



import os
import tempfile
import unittest
from geogeometry import CadReader


class TestPlot3d(unittest.TestCase):

    def setUp(self):
        self.base_dir = os.path.dirname(os.path.abspath(__file__))
        self.resource_dir = os.path.join(self.base_dir, '..', 'test_io', 'resources', 'dxf_files')

        self.file_006_path = os.path.join(self.resource_dir, "file_006.dxf")  # Mixed

    def test_file_006_V01(self):
        models_collection = CadReader.readFile(self.file_006_path)

        tmpfile = tempfile.NamedTemporaryFile(suffix=".png", delete=False)
        models_collection.plot2d(filepath=tmpfile.name)

        self.assertTrue(os.path.exists(tmpfile.name))

        tmpfile.close()
        if os.path.exists(tmpfile.name):
            os.remove(tmpfile.name)

    def test_file_006_V02(self):
        models_collection = CadReader.readFile(self.file_006_path)

        for plane in ['xy', 'yz', 'xz']:
            tmpfile = tempfile.NamedTemporaryFile(suffix=".png", delete=False)
            models_collection.plot2d(filepath=tmpfile.name, projection_mode="3d", projection_plane=plane)

            self.assertTrue(os.path.exists(tmpfile.name))

            tmpfile.close()
            if os.path.exists(tmpfile.name):
                os.remove(tmpfile.name)


if __name__ == "__main__":
    unittest.main()



from typing import TYPE_CHECKING

from geogeometry.geometry.shared.explicit_geometry.ExplicitGeometryTransformer import ExplicitGeometryTransformer

if TYPE_CHECKING:
    from geogeometry.geometry.polyline.Polyline import Polyline


class PolylineTransformer(ExplicitGeometryTransformer):

    def __init__(self, polyline: 'Polyline'):
        super().__init__(element=polyline)



from typing import Optional, TYPE_CHECKING, Union, List, Literal

import numpy as np

from geogeometry.geometry.points.collection_components.PointsCollectionPlotter import PointsCollectionPlotter
from geogeometry.geometry.shared.BaseObjectCollection import BaseObjectCollection


class PointsCollection(BaseObjectCollection['PointsCollection', np.ndarray]):

    def __init__(self, name: Optional[str] = None, points: Optional[Union[List, np.ndarray]] = None):
        super().__init__(name=name)

        if points is not None:
            self.setPoints(points=points)

        self.plotter: PointsCollectionPlotter = PointsCollectionPlotter(self)

    def setPoints(self, points: Union[List, np.ndarray]) -> None:
        if isinstance(points, list):
            points = np.array(points)
        self.elements = points

    def getPoints(self) -> Union[List, np.ndarray]:
        return self.elements



class PolylinePlotProperties(object):

    def __init__(self):
        super().__init__()

        # 2D
        self.line_width: float = 2.
        self.line_style: str = '-'

        # 3D
        self.tube_radius: float = 1.

    def setLineWidth(self, line_width: float) -> None:
        self.line_width = line_width

    def setLineStyle(self, line_style: str) -> None:
        self.line_style = line_style

    def setTubeRadius(self, tube_radius: float) -> None:
        self.tube_radius = tube_radius

    def getLineWidth(self) -> float:
        return self.line_width

    def getLineStyle(self) -> str:
        return self.line_style

    def getTubeRadius(self) -> float:
        return self.tube_radius



from typing import TYPE_CHECKING, Optional

import pyvista as pv

from geogeometry.geometry.shared.shared_components.BasePlotter import BasePlotter

if TYPE_CHECKING:
    import matplotlib.pyplot as plt
    from geogeometry.geometry.model.GeometryModel import GeometryModel


class GeometryModelPlotter(BasePlotter):

    def __init__(self, model: 'GeometryModel'):
        super().__init__(element=model)
        self.model: 'GeometryModel' = model

    def addToMatplotlibAx(self, ax: 'plt.Axes') -> None:
        if len(self.model.getPolylines()):
            self.model.getPolylines().addToMatplotlibAx(ax=ax)

        if len(self.model.getTriangulations()):
            self.model.getTriangulations().addToMatplotlibAx(ax=ax)

        if len(self.model.getPoints()):
            self.model.getPoints().addToMatplotlibAx(ax=ax)

    def addToPyvistaPlotter(self, plotter: 'pv.Plotter', **kwargs) -> None:
        if len(self.model.getPolylines()):
            self.model.getPolylines().addToPyvistaPlotter(plotter=plotter, **kwargs)

        if len(self.model.getTriangulations()):
            self.model.getTriangulations().addToPyvistaPlotter(plotter=plotter, **kwargs)

        if len(self.model.getPoints()):
            self.model.getPoints().addToPyvistaPlotter(plotter=plotter, **kwargs)

    def addToParaviewPlotter(self, plotter, **kwargs) -> None: ...



from typing import Optional, Literal, TYPE_CHECKING

import numpy as np

from geogeometry.geometry.shared.shared_components.BaseMetrics import BaseMetrics

if TYPE_CHECKING:
    from geogeometry.geometry.shared.explicit_geometry.ExplicitGeometry import ExplicitGeometry


class ExplicitGeometryMetrics(BaseMetrics):

    def __init__(self, element: 'ExplicitGeometry'):
        self.element: 'ExplicitGeometry' = element

        # Metrics
        self.dimensions: Optional[Literal[2, 3]] = None
        self.limits: Optional[np.ndarray] = None
        self.center: Optional[np.ndarray] = None

    def getDimensions(self) -> Optional[Literal[2, 3]]:
        return self.dimensions

    def getLimits(self) -> Optional[np.ndarray]:
        return self.limits

    def getCenter(self) -> Optional[np.ndarray]:
        return self.center

    def calculateMetrics(self) -> None:
        self.calculateDimensions()
        self.calculateLimits()
        self.calculateCenter()

    def calculateDimensions(self) -> None:
        self.dimensions = self.element.getNodes().shape[1]

    def calculateLimits(self) -> None:
        _min, _max = np.min(np.array(self.element.getNodes()), axis=0), np.max(np.array(self.element.getNodes()), axis=0)
        self.limits = np.array([_min, _max])

    def calculateCenter(self) -> None:
        self.center = np.average(self.limits, axis=0)



import math
from typing import Union, List, TYPE_CHECKING

import numpy as np

from geogeometry.geometry.operations.Angles import Angles

if TYPE_CHECKING:
    from geogeometry.geometry.vector.Vector import Vector


class Rotations(object):

    @staticmethod
    def calculate2dRotationMatrix(angle: float) -> np.ndarray:
        """
        Return anti-clockwise 2d rotation matrix.
        :param angle: (float) radians
        :return: 2D rotation matrix
        """
        rot_m = [
            [math.cos(angle), -math.sin(angle)],
            [math.sin(angle), math.cos(angle)],
        ]

        return np.array(rot_m)

    @staticmethod
    def calculateXRotationMatrix(angle: float) -> np.ndarray:
        """
        Returns anti-clockwise x-axis rotation matrix for 'angle'.
        :param angle: radians.
        :return: Rotation Matrix.
        """
        rot_x = [
            [1, 0, 0],
            [0, math.cos(angle), -math.sin(angle)],
            [0, math.sin(angle), math.cos(angle)]
        ]
        return np.array(rot_x)

    @staticmethod
    def calculateYRotationMatrix(angle):
        """
        Returns anti-clockwise y-axis rotation matrix for 'angle'.
        :param angle: radians.
        :return: Rotation Matrix.
        """
        rot_y = [
            [math.cos(angle), 0, -math.sin(angle)],
            [0, 1, 0],
            [math.sin(angle), 0, math.cos(angle)]
        ]
        return np.array(rot_y)

    @staticmethod
    def calculateZRotationMatrix(angle: float) -> np.ndarray:
        """
        Returns anti-clockwise z-axis rotation matrix for 'angle'.
        :param angle: radians.
        :return: Rotation Matrix.
        """
        rot_z = [
            [math.cos(angle), -math.sin(angle), 0],
            [math.sin(angle), math.cos(angle), 0],
            [0, 0, 1]
        ]
        return np.array(rot_z)

    @staticmethod
    def rotate2dPositionVector(vector: 'Vector', angle: float) -> np.ndarray:
        """
        Rotates position 'vector' (origin=(0,0,0)) anti-clockwise by 'angle'.
        vector: (x, y)
        angle: float (radians)
        """
        if (vector.getDimensions() == 3) or (not vector.isPositionVector()):
            raise ValueError("Vector must be 2D and positional.")

        rot_matrix = Rotations.calculate2dRotationMatrix(angle=angle)

        return rot_matrix.dot(vector.getTip())

    @staticmethod
    def rotatePointsByRotationMatrix(points: Union[List, np.ndarray], rotation_matrix: np.ndarray) -> np.ndarray:
        """
        Rotates 'points' by previously defined 'rotation_matrix'.
        :param points:
        :param rotation_matrix:
        :return:
        """
        if isinstance(points, list):
            points = np.array(points)
        rotated_points = rotation_matrix.dot(points.T).T
        rotated_points = np.round(rotated_points, decimals=5)  # Rounded because of small np tolerance (e-17)
        return rotated_points

    @staticmethod
    def calculateRotationMatrixFromVectors(v0: 'Vector', v1: 'Vector') -> np.ndarray:

        # if v0.getNormalizedVector() == v1.getNormalizedVector():
        if v0.isVectorParallel(other_vector=v1):
            return np.identity(3)

        angle = v0.calculateAngleWithVector(other_vector=v1)
        if abs(angle) == 180.:
            return -np.identity(3)

        # pos0, pos1 = v0.getNormalizedVector().getTip(), v1.getNormalizedVector().getTip()
        pos0, pos1 = v0.getUnitVector(), v1.getUnitVector()

        # Rotation axis
        v = np.cross(pos0, pos1)
        v /= np.linalg.norm(v)

        s = np.linalg.norm(v)
        c = np.dot(pos0, pos1)
        I = np.eye(3)

        vx = np.array([[0, -v[2], v[1]],
                       [v[2], 0, -v[0]],
                       [-v[1], v[0], 0]])

        v2 = np.matmul(vx, vx)

        R = I + vx + v2 * ((1. - c) / (s ** 2))

        if np.linalg.det(R) < 0:
            R[:, 0] = -R[:, 0]  # Flip one column to ensure a valid right-handed system

        return np.array(R)



from typing import Optional, Union, List

import numpy as np
import trimesh


class TriangulationProperties(object):

    def __init__(self):

        # self.nodes: Optional[np.ndarray] = None
        self.faces: Optional[np.ndarray] = None
        self.markers: Optional[np.ndarray] = None

        self.trimesh_t: Optional[trimesh.base.Trimesh] = None

    # def setNodes(self, nodes: Union[List, np.ndarray]) -> None:
    #     if isinstance(nodes, list):
    #         nodes = np.array(nodes)
    #     self.nodes = nodes

    def setFaces(self, faces: Union[List[int], np.ndarray[int]]) -> None:
        if isinstance(faces, list):
            faces = np.array(faces).astype(int)
        self.faces = faces

    def setMarkers(self, markers: Union[List[int], np.ndarray[int]]) -> None:
        if isinstance(markers, list):
            markers = np.array(markers).astype(int)
        self.markers = markers

    def setTrimesh(self, trimesh_t: trimesh.base.Trimesh) -> None:
        self.trimesh_t = trimesh_t
        # self.nodes = None
        self.faces = None

    # def getNodes(self) -> Optional[np.ndarray]:
    #     if self.trimesh_t is not None:
    #         return np.asarray(self.trimesh_t.vertices)
    #     else:
    #         return self.nodes

    def getFaces(self) -> Optional[np.ndarray]:
        if self.trimesh_t is not None:
            return np.asarray(self.trimesh_t.faces)
        else:
            return self.faces

    def getMarkers(self) -> Optional[np.ndarray]:
        return self.markers

    def getTrimesh(self) -> Optional[trimesh.base.Trimesh]:
        return self.trimesh_t



from typing import Optional, TYPE_CHECKING

import numpy as np

from geogeometry.geometry.shared.shared_components.BaseMetrics import BaseMetrics

if TYPE_CHECKING:
    from geogeometry.geometry.circle.Circle import Circle
    from geogeometry.geometry.vector.Vector import Vector


class CircleMetrics(BaseMetrics):

    def __init__(self, circle: 'Circle'):
        self.circle: 'Circle' = circle

        self.area: Optional[float] = None
        self.normal_vector: Optional['Vector'] = None

    def calculateMetrics(self) -> None:
        self.calculateArea()
        self.calculateNormalVector()

    def calculateArea(self) -> None:
        self.area = np.pi * (self.circle.getRadius() ** 2)

    def calculateNormalVector(self) -> None:
        if self.circle.getPlane() is None:
            self.normal_vector = Vector(tip=[0, 0, 1])
        else:
            self.normal_vector = self.circle.getPlane().getNormalVector()



from typing import List, Union, TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from geogeometry.geometry.shared.BaseObject import BaseObject
    from geogeometry.geometry.shared.BaseObjectCollection import BaseObjectCollection


class PlotFigure(object):

    def __init__(self):
        self.elements: List[Union['BaseObject', 'BaseObjectCollection']] = []

    def addElement(self, element: Union['BaseObject', 'BaseObjectCollection']) -> None:
        self.elements += [element]

    def calculateLimits(self) -> np.ndarray:
        all_limits = [lim for e in self.elements for lim in e.getLimits()]
        limits = np.array([np.min(all_limits, axis=0), np.max(all_limits, axis=0)])
        return limits



from typing import List, TypeVar, Iterator, Optional, Union, Generic, TYPE_CHECKING

import numpy as np

from geogeometry.geometry.shared.BaseObject import BaseObject
from geogeometry.geometry.shared.observer.Observer import Observer
from geogeometry.geometry.shared.shared_components.BaseCollectionMetrics import BaseCollectionMetrics
from geogeometry.geometry.shared.shared_components.BaseCollectionPlotter import BaseCollectionPlotter

if TYPE_CHECKING:
    from geogeometry.geometry.shared.observer.Observable import Observable

CollectionType = TypeVar('CollectionType')
ElementType = TypeVar('ElementType', bound='BaseObject')


class BaseObjectCollection(BaseObject[CollectionType], Observer, Generic[CollectionType, ElementType]):

    def __init__(self, name: Optional[str] = None):
        super().__init__(name=name)

        self.elements: List[ElementType] = []

        self.metrics: BaseCollectionMetrics = BaseCollectionMetrics(self)
        self.plotter: BaseCollectionPlotter = BaseCollectionPlotter(self)

    def __len__(self) -> int:
        return len(self.elements)

    def __iter__(self) -> Iterator[ElementType]:
        for e in self.elements:
            yield e

    def __getitem__(self, identifier: Union[int, str]) -> ElementType:
        if isinstance(identifier, int):
            return self.elements[identifier]
        else:
            for e in self.elements:
                if e.getName() == identifier:
                    return e
            else:
                raise ValueError(f"Element '{identifier}' not found in collection.")

    def addElement(self, element: ElementType) -> None:
        self.elements += [element]
        element.addObserver(observer=self)

        self.notifyObservers()
        self.calculateMetrics()

    def deleteElement(self, identifier: Union[int, str]):
        for e in self.elements:
            if e.getName() == identifier:
                self.elements.remove(e)
                e.removeObserver(observer=self)

                self.notifyObservers()
                self.calculateMetrics()
                break
        else:
            raise ValueError(f"Element '{identifier}' not found in collection.")

    def getLimits(self) -> Optional[np.ndarray]:
        return self.metrics.getLimits()

    def getCenter(self) -> Optional[np.ndarray]:
        return self.metrics.getLimits()

    def onElementChange(self, observable: 'Observable'):
        self.calculateMetrics()
        self.notifyObservers()



from typing import Optional, TYPE_CHECKING, Literal, Callable, Union, List

import numpy as np
import trimesh

from geogeometry.geometry.shared.BaseObject import BaseObject
from geogeometry.geometry.shared.explicit_geometry.ExplicitGeometry import ExplicitGeometry
from geogeometry.geometry.triangulation.components.TriangulationMetrics import TriangulationMetrics
from geogeometry.geometry.triangulation.components.TriangulationPlotter import TriangulationPlotter
from geogeometry.geometry.triangulation.components.TriangulationProperties import TriangulationProperties


def processTrimeshAndMetrics(func: Callable[..., None]) -> Callable[..., None]:

    def inner(triangulation: 'Triangulation', *args, **kwargs) -> None:
        if triangulation.trimesh_t is not None:
            raise ValueError("Triangulation already set.")

        func(triangulation, *args, **kwargs)

        if triangulation.getNodes() is not None and triangulation.getFaces() is not None:
            trimesh_t = trimesh.Trimesh(vertices=triangulation.getNodes(), faces=triangulation.getFaces(), process=True)
            triangulation.setTrimesh(trimesh_t=trimesh_t)

            triangulation.resetNodes()  # Nodes within trimesh now

            triangulation.metrics.calculateMetrics()

    return inner


class Triangulation(ExplicitGeometry, TriangulationProperties):

    def __init__(self, name: Optional[str] = None):
        super().__init__(name=name)

        self.metrics: TriangulationMetrics = TriangulationMetrics(self)
        self.plotter: TriangulationPlotter = TriangulationPlotter(self)

    def __len__(self) -> int:
        if self.getFaces() is None:
            return 0
        return len(self.getFaces())

    def __str__(self) -> str:
        txt = f"Triangle(name={self.getName()}, faces={len(self)})"
        return txt

    def getNodes(self) -> Optional[np.ndarray]:
        if self.getTrimesh() is not None:
            return np.asarray(self.getTrimesh().vertices)
        else:
            return super().getNodes()

    @processTrimeshAndMetrics
    def setNodes(self, nodes: Union[List, np.ndarray]) -> None:
        super().setNodes(nodes=nodes)

    @processTrimeshAndMetrics
    def setFaces(self, faces: Union[List[int], np.ndarray[int]]) -> None:
        TriangulationProperties.setFaces(self, faces=faces)



from typing import Optional, TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from geogeometry.geometry.vector.Vector import Vector


class CylinderProperties(object):

    def __init__(self):
        self.radius: Optional[float] = None
        self.origin: Optional[np.ndarray] = None
        self.axis_vector: Optional['Vector'] = None
        self.length: Optional[float] = None

    def setRadius(self, radius: float) -> None:
        self.radius = radius

    def setOrigin(self, origin: np.ndarray) -> None:
        self.origin = origin

    def setAxisVector(self, axis_vector: 'Vector'):
        self.axis_vector = axis_vector

    def setLength(self, length: float) -> None:
        self.length = length

    def getRadius(self) -> float:
        return self.radius

    def getOrigin(self) -> np.ndarray:
        return self.origin

    def getAxisVector(self) -> 'Vector':
        return self.axis_vector

    def getLength(self) -> float:
        return self.length



from typing import Optional, TYPE_CHECKING, Literal

from geogeometry.geometry.polyline.collection_components.PolylinesCollectionPlotter import PolylinesCollectionPlotter
from geogeometry.geometry.shared.BaseObjectCollection import BaseObjectCollection

if TYPE_CHECKING:
    from geogeometry.geometry.polyline.Polyline import Polyline


class PolylinesCollection(BaseObjectCollection['PolylinesCollection', 'Polyline']):

    def __init__(self, name: Optional[str] = None):
        super().__init__(name=name)

        # self.plotter: PolylinesCollectionPlotter = PolylinesCollectionPlotter(self)

    def __str__(self) -> str:
        txt = f"PolylinesCollection(name={self.getName()}, polylines={len(self)})"
        return txt



from typing import Optional, TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from geogeometry.geometry.vector.Vector import Vector


class PlaneProperties(object):

    def __init__(self):

        self.origin: Optional[np.ndarray] = None
        self.normal_vector: Optional['Vector'] = None

    def setOrigin(self, origin: np.ndarray) -> None:
        self.origin = origin

    def setNormalVector(self, normal_vector: 'Vector') -> None:
        self.normal_vector = normal_vector

    def getOrigin(self) -> Optional[np.ndarray]:
        return self.origin

    def getNormalVector(self) -> Optional['Vector']:
        return self.normal_vector



import os
import unittest
from geogeometry import CadReader


class TestReadRhinoTriangulations(unittest.TestCase):

    def setUp(self):
        self.base_dir = os.path.dirname(os.path.abspath(__file__))
        self.resource_dir = os.path.join(self.base_dir, 'resources', 'stl_files')

        self.extension = "stl"

        self.file_001_path = os.path.join(self.resource_dir, "file_001." + self.extension)

    def test_file_001(self):
        models_collection = CadReader.readFile(self.file_001_path)
        self.assertEqual(len(models_collection), 1)

        n_triangulations = len([m.getTriangulations() for m in models_collection])
        self.assertEqual(n_triangulations, 1)


if __name__ == "__main__":
    unittest.main()



from abc import ABC, abstractmethod
from typing import Optional, Literal, Union, TYPE_CHECKING

from geogeometry.graphics.Figure2d import Figure2d
from geogeometry.graphics.Figure3d import Figure3d

if TYPE_CHECKING:
    import pyvista as pv
    import matplotlib.pyplot as plt
    from geogeometry.geometry.shared.BaseObject import BaseObject
    from geogeometry.geometry.shared.BaseObjectCollection import BaseObjectCollection


class BasePlotter(ABC):

    def __init__(self, element: Union['BaseObject', 'BaseObjectCollection']):
        self.element: Union['BaseObject', 'BaseObjectCollection'] = element

    def addToFigure2d(self, figure_2d: 'Figure2d') -> None:
        figure_2d.addElement(self.element)

    @abstractmethod
    def addToMatplotlibAx(self, ax: 'plt.Axes') -> None: ...

    def addToFigure3d(self, figure_3d: 'Figure3d') -> None:
        figure_3d.addElement(self.element)

    @abstractmethod
    def addToPyvistaPlotter(self, plotter: 'pv.Plotter', **kwargs) -> None: ...

    @abstractmethod
    def addToParaviewPlotter(self, plotter, **kwargs) -> None: ...

    def plot2d(self, filepath: Optional[str] = None,
               projection_mode: Literal['2d', '3d'] = '2d',
               projection_plane: Literal['xy', 'yz', 'xz'] = 'xy',
               **kwargs) -> None:

        figure_2d = Figure2d()
        self.addToFigure2d(figure_2d=figure_2d)
        figure_2d.plot(filepath=filepath, projection_mode=projection_mode, projection_plane=projection_plane)

    def plot3d(self, engine: Literal['pyvista', 'paraview'] = 'pyvista',
               filepath: Optional[str] = None,
               projection_plane: Optional[Literal['xy', 'yz', 'xz']] = None,
               show_axes: bool = True,
               **kwargs) -> None:

        figure_3d = Figure3d(engine=engine)
        self.addToFigure3d(figure_3d=figure_3d)
        figure_3d.plot(filepath=filepath, projection_plane=projection_plane, show_axes=show_axes)



from typing import Optional, Union

import ezdxf
import ezdxf.entities

from geogeometry.geo_io.readers.dxf_components.EntityHandlerRegistry import EntityHandlerRegistry

from geogeometry.geometry.model.GeometryModel import GeometryModel
from geogeometry.geometry.model.GeometryModelsCollection import GeometryModelsCollection

from geogeometry.geometry.polyline.Polyline import Polyline
from geogeometry.geometry.triangulation.Triangulation import Triangulation


class DxfReader(object):

    @staticmethod
    def readFile(filepath: str) -> Optional[GeometryModelsCollection]:
        doc = ezdxf.readfile(filepath)
        msp = doc.modelspace()
        groups = msp.groupby(dxfattrib="layer")

        handlers_registry = EntityHandlerRegistry()

        models_collection = GeometryModelsCollection()
        for layer, entities in groups.items():

            model = GeometryModel(name=layer)

            for entity in entities:
                handler = handlers_registry.getHandler(entity.dxftype())

                if handler:
                    result = handler(entity)
                    if isinstance(result, Polyline):
                        model.addPolyline(result)
                    elif isinstance(result, Triangulation):
                        model.addTriangulation(result)
                else:
                    raise ValueError(f"Entity type '{entity.dxftype()}' not supported.")

            if len(model):
                models_collection.addElement(model)

        return models_collection



from typing import Optional, Union, List, Literal

import numpy as np

from geogeometry.geometry.shared.BaseObject import BaseObject

from geogeometry.geometry.shared.explicit_geometry.ExplicitGeometryProperties import ExplicitGeometryProperties
from geogeometry.geometry.shared.explicit_geometry.ExplicitGeometryMetrics import ExplicitGeometryMetrics
from geogeometry.geometry.shared.explicit_geometry.ExplicitGeometryTransformer import ExplicitGeometryTransformer


class ExplicitGeometry(BaseObject, ExplicitGeometryProperties):

    def __init__(self, name: Optional[str] = None):
        super().__init__(name=name)

        self.explicit_metrics: ExplicitGeometryMetrics = ExplicitGeometryMetrics(self)
        self.explicit_transformer: ExplicitGeometryTransformer = ExplicitGeometryTransformer(self)

    def calculateMetrics(self) -> None:
        self.explicit_metrics.calculateMetrics()  # Basic explicit metrics
        self.metrics.calculateMetrics()  # Custom defined metrics

    def translate(self, translation_vector: Union[List, np.ndarray]) -> None:
        self.explicit_transformer.translate(translation_vector=translation_vector)

    def rotateByRotationMatrix(self, rotation_matrix: np.ndarray) -> None:
        self.explicit_transformer.rotateByRotationMatrix(rotation_matrix=rotation_matrix)



from typing import TYPE_CHECKING

import numpy as np

from geogeometry.geometry.shared.explicit_geometry.ExplicitGeometryTransformer import ExplicitGeometryTransformer, updateElementMetrics

if TYPE_CHECKING:
    from geogeometry.geometry.vector.Vector import Vector


class VectorTransformer(ExplicitGeometryTransformer):

    def __init__(self, vector: 'Vector'):
        super().__init__(element=vector)
        self.vector: 'Vector' = vector

    @updateElementMetrics
    def reverse(self) -> None:
        n0 = -1 * self.vector.getOrigin()
        n1 = -1 * self.vector.getTip()
        self.vector.setNodes(nodes=np.array([n0, n1]))



from geogeometry.geometry.triangulation.Triangulation import Triangulation
from geogeometry.geometry.polyline.Polyline import Polyline
from geogeometry.geometry.vector.Vector import Vector
from geogeometry.geometry.plane.Plane import Plane

from geogeometry.geometry.polyline.PolylinesCollection import PolylinesCollection

from geogeometry.geo_io.CadReader import CadReader

from geogeometry.geometry.operations.Angles import Angles
from geogeometry.geometry.operations.Rotations import Rotations



import os
from typing import Optional, TYPE_CHECKING

from geogeometry.geo_io.readers.CsvReader import CsvReader
from geogeometry.geo_io.readers.RhinoReader import RhinoReader
from geogeometry.geo_io.readers.StlReader import StlReader
from geogeometry.geo_io.readers.DxfReader import DxfReader

if TYPE_CHECKING:
    from geogeometry.geometry.model.GeometryModelsCollection import GeometryModelsCollection


class CadReader(object):

    @staticmethod
    def readFile(filepath: str, **kwargs) -> Optional['GeometryModelsCollection']:

        if not os.path.isfile(filepath):
            raise ValueError(f"File '{filepath}' not found.")

        extension = filepath.split('.')[-1].lower()

        if extension == 'dxf':
            return DxfReader.readFile(filepath)
        elif extension == '3dm':
            return RhinoReader.readFile(filepath)
        elif extension == 'stl':
            return StlReader.readFile(filepath)
        elif extension == 'csv':
            return CsvReader.readFile(filepath, **kwargs)
        else:
            raise ValueError(f"File extension '.{extension}' not supported.")



from typing import TYPE_CHECKING

from geogeometry.geometry.shared.explicit_geometry.ExplicitGeometryMetrics import ExplicitGeometryMetrics
from geogeometry.geometry.shared.shared_components.BaseMetrics import BaseMetrics

if TYPE_CHECKING:
    from geogeometry.geometry.triangulation.Triangulation import Triangulation


class TriangulationMetrics(BaseMetrics):

    def __init__(self, triangulation: 'Triangulation'):
        self.triangulation: 'Triangulation' = triangulation

    def calculateMetrics(self) -> None:
        pass



from typing import TYPE_CHECKING, List

if TYPE_CHECKING:
    from geogeometry.geometry.shared.observer.Observer import Observer


class Observable(object):

    def __init__(self):
        super().__init__()
        self.observers: List['Observer'] = []

    def addObserver(self, observer: 'Observer'):
        self.observers += [observer]

    def removeObserver(self, observer: 'Observer'):
        self.observers.remove(observer)

    def notifyObservers(self):
        for observer in self.observers:
            observer.onElementChange(self)



from typing import Optional, Literal

import numpy as np

from geogeometry.geometry.shared.BasePlotProperties import BasePlotProperties


class BaseObjectProperties(BasePlotProperties):

    def __init__(self, name: Optional[str] = None):
        super().__init__()

        self.name: Optional[str] = name
        self.id: int = id(self)

        # self.plot_properties: BasePlotProperties = BasePlotProperties()

        # Metrics
        # self.dimensions: Optional[Literal[2, 3]] = None
        # self.limits: Optional[np.ndarray] = None
        # self.centroid: Optional[np.ndarray] = None

        # Plot properties
        # self.color: str = 'red'
        # self.opacity: float = 1.0

    def setName(self, name: str) -> None:
        self.name = name

    def setId(self, _id: int) -> None:
        self.id = _id

    # def setDimensions(self, dimensions: Literal[2, 3]) -> None:
    #     self.dimensions = dimensions
    #
    # def setLimits(self, limits: np.ndarray) -> None:
    #     self.limits = limits
    #
    # def setCentroid(self, centroid: np.ndarray) -> None:
    #     self.centroid = centroid

    # def setColor(self, color: str) -> None:
    #     self.color = color
    #
    # def setOpacity(self, opacity: float) -> None:
    #     self.opacity = opacity

    def getName(self) -> str:
        return self.name

    def getId(self) -> int:
        return self.id

    # def getDimensions(self) -> Optional[Literal[2, 3]]:
    #     return self.dimensions
    #
    # def getLimits(self) -> Optional[np.ndarray]:
    #     return self.limits
    #
    # def getCentroid(self) -> Optional[np.ndarray]:
    #     return self.centroid

    # def getColor(self) -> str:
    #     return self.color
    #
    # def getOpacity(self) -> float:
    #     return self.opacity



from typing import TYPE_CHECKING, Optional

import numpy as np

from geogeometry.geometry.operations.Angles import Angles
from geogeometry.geometry.shared.shared_components.BaseMetrics import BaseMetrics

if TYPE_CHECKING:
    from geogeometry.geometry.vector.Vector import Vector


class VectorMetrics(BaseMetrics):

    def __init__(self, vector: 'Vector'):
        self.vector: 'Vector' = vector

        self.length: Optional[float] = None

        self.is_unitary: Optional[bool] = None
        self.is_position_vector: Optional[bool] = None  # origin=(0,0,0)

        self.unit_vector: Optional[np.ndarray] = None

        self.dip: Optional[float] = None
        self.azimuth: Optional[float] = None

    def getLength(self) -> float:
        return self.length

    def isUnitary(self) -> bool:
        return self.is_unitary

    def isPositionVector(self) -> bool:
        return self.is_position_vector

    def getUnitVector(self) -> np.ndarray:
        return self.unit_vector

    def getDip(self) -> float:
        return self.dip

    def getAzimuth(self) -> float:
        return self.azimuth

    def calculateMetrics(self) -> None:
        self.calculateLength()

        self.is_unitary = (self.length == 1.)
        self.is_position_vector = (np.sum(self.vector.getOrigin()) == 0.)

        self.calculateUnitVector()
        self.calculateDipAndAzimuth()

    def calculateLength(self) -> None:
        diff = self.vector.getTip() - self.vector.getOrigin()
        self.length = np.linalg.norm(diff)

    def calculateUnitVector(self) -> None:
        if self.is_unitary and self.is_position_vector:
            self.unit_vector = self.vector.getTip()
            # self.vector.setNormalizedVector(normalized_vector=self.vector.copy())
            return

        pos = self.vector.getTip() - self.vector.getOrigin()
        norm = np.linalg.norm(pos)
        if norm == 0.:
            raise ValueError("Origin as vector.")

        self.unit_vector = np.round(pos/norm, decimals=5)

    def calculateDipAndAzimuth(self) -> None:

        # if self.vector.getNormalizedVector() is None:
        #     self.calculateNormalizedVector()

        # norm_vector = self.vector.getNormalizedVector()

        azimuth = None

        if self.vector.getDimensions() == 2:
            dip = 0.
        # if norm_vector.getOrigin().shape[0] == 2:
        #     dip = 90.
        else:
            # if abs(norm_vector.getTip()[2]) == 1.:
            if abs(self.unit_vector[2]) == 1.:
                dip, azimuth = 0, 0
            else:
                dip = Angles.calculateAngleBetweenVectorAndAxis(v=self.vector, axis_id='z')
                if dip > 90.:
                    reversed_vector = self.vector.copy()
                    reversed_vector.reverse()
                    dip = Angles.calculateAngleBetweenVectorAndAxis(v=reversed_vector, axis_id='z')

        if azimuth is None:
            azimuth = Angles.calculateAngleFromThreePoints([0., 1.], [0., 0.], self.vector.getTip()[:2])

        if self.vector.getTip()[0] < 0.:
            azimuth = 360. - azimuth

        self.dip = round(dip, 2)
        self.azimuth = round(azimuth, 2)



import copy
import gzip
import lzma
import pickle

from typing import Optional, TYPE_CHECKING, Literal, Generic, TypeVar

from geogeometry.geometry.shared.BaseObjectProperties import BaseObjectProperties
from geogeometry.geometry.shared.observer.Observable import Observable


if TYPE_CHECKING:
    import pyvista as pv
    import matplotlib.pyplot as plt
    from geogeometry.graphics.Figure2d import Figure2d
    from geogeometry.graphics.Figure3d import Figure3d
    from geogeometry.geometry.shared.shared_components.BaseMetrics import BaseMetrics
    from geogeometry.geometry.shared.shared_components.BasePlotter import BasePlotter


ObjectType = TypeVar('ObjectType')


class BaseObject(BaseObjectProperties, Observable, Generic[ObjectType]):

    def __init__(self, name: Optional[str] = None):
        super().__init__(name=name)

        # These components must be defined on child classes (emuling "abstract fields")
        self.metrics: Optional['BaseMetrics'] = None
        self.plotter: Optional['BasePlotter'] = None

    def copy(self):
        copy_p = copy.deepcopy(self)
        copy_p.setId(_id=id(copy_p))
        return copy_p

    # IO
    def save(self, savepath: str) -> None:
        if '.pkl.gz' in savepath:
            pickle.dump(self, gzip.open(savepath, 'wb'), pickle.HIGHEST_PROTOCOL)
        elif 'pkl.xz' in savepath:
            pickle.dump(self, lzma.open(savepath, 'wb'), pickle.HIGHEST_PROTOCOL)
        else:
            pickle.dump(self, open(savepath, 'wb'), pickle.HIGHEST_PROTOCOL)

        print(f'{self.__class__.__name__} instance saved at: "' + savepath + '"')

    @staticmethod
    def load(load_path) -> ObjectType:
        if '.pkl.gz' in load_path:
            return pickle.load(gzip.open(load_path, 'rb'))
        elif 'pkl.xz' in load_path:
            return pickle.load(lzma.open(load_path, 'rb'))
        else:
            return pickle.load(open(load_path, 'rb'))

    def calculateMetrics(self) -> None:
        self.metrics.calculateMetrics()

    def addToFigure2d(self, figure_2d: 'Figure2d') -> None:
        self.plotter.addToFigure2d(figure_2d=figure_2d)

    def addToMatplotlibAx(self, ax: 'plt.Axes') -> None:
        self.plotter.addToMatplotlibAx(ax=ax)

    def addToFigure3d(self, figure_3d: 'Figure3d') -> None:
        self.plotter.addToFigure3d(figure_3d=figure_3d)

    def addToPyvistaPlotter(self, plotter: 'pv.Plotter', **kwargs) -> None:
        self.plotter.addToPyvistaPlotter(plotter=plotter)

    def addToParaviewPlotter(self, plotter, **kwargs) -> None:
        self.plotter.addToParaviewPlotter(plotter=plotter, **kwargs)

    def plot2d(self, filepath: Optional[str] = None, projection_mode: Literal['2d', '3d'] = '2d', **kwargs) -> None:
        self.plotter.plot2d(filepath=filepath, projection_mode=projection_mode, **kwargs)

    def plot3d(self, engine: Literal['pyvista', 'paraview'] = 'pyvista',
               filepath: Optional[str] = None,
               projection_plane: Optional[Literal['xy', 'yz', 'xz']] = None,
               show_axes: bool = True,
               **kwargs) -> None:
        self.plotter.plot3d(engine=engine, filepath=filepath,
                            projection_plane=projection_plane, show_axes=show_axes, **kwargs)





from typing import Union, List, Callable

import numpy as np

from geogeometry.geometry.operations.Rotations import Rotations


def updateElementMetrics(func: Callable[..., None]) -> Callable[..., None]:

    def inner(transformer, *args, **kwargs) -> None:
        func(transformer, *args, **kwargs)
        transformer.element.calculateMetrics()

    return inner


class ExplicitGeometryTransformer(object):

    def __init__(self, element):
        self.element = element

    @updateElementMetrics
    def translate(self, translation_vector: Union[List, np.ndarray]) -> None:
        self.element.setNodes(nodes=self.element.getNodes() + translation_vector)

    @updateElementMetrics
    def rotateByRotationMatrix(self, rotation_matrix: np.ndarray) -> None:
        rotated_nodes = Rotations.rotatePointsByRotationMatrix(points=self.element.getNodes(), rotation_matrix=rotation_matrix)
        self.element.setNodes(nodes=rotated_nodes)



from typing import Optional, Union

import ezdxf
import ezdxf.entities

import numpy as np
from ezdxf.math import Vec3, Matrix44

from geogeometry.geometry.operations.Rotations import Rotations
from geogeometry.geometry.triangulation.Triangulation import Triangulation
from geogeometry.geometry.polyline.Polyline import Polyline
from geogeometry.geometry.vector.Vector import Vector


class DxfEntityHandler(object):

    @staticmethod
    def handlePolyline(entity: ezdxf.entities.Polyline) -> Optional[Union[Triangulation, Polyline]]:

        if entity.is_poly_face_mesh:
            return DxfEntityHandler.handlePolyfaceMesh(entity=entity)
        else:
            return DxfEntityHandler.handleActualPolyline(entity=entity)

    @staticmethod
    def handlePolyfaceMesh(entity: ezdxf.entities.Polyline) -> Optional[Triangulation]:

        vertices = []
        faces = []
        for vertex in entity.vertices:
            if vertex.is_face_record:
                # Just considers triangulations
                indices = [vertex.get_dxf_attrib(name, 0) for name in ('vtx0', 'vtx1', 'vtx2')]
                # indices = [vertex.dxf.vtx0, vertex.dxf.vtx1, vertex.dxf.vtx2, vertex.dxf.vtx3]

                indices = [idx - 1 for idx in indices if idx != 0]
                faces += [indices]

            elif vertex.is_poly_face_mesh_vertex:
                x, y, z = vertex.dxf.location
                vertices += [[x, y, z]]

        vertices = np.array(vertices)

        t = Triangulation()
        t.setNodes(nodes=vertices)
        t.setFaces(faces=faces)

        return t

    @staticmethod
    def handleActualPolyline(entity: ezdxf.entities.Polyline) -> Optional[Polyline]:
        plane_normal = Vector(tip=np.array(entity.dxf.extrusion))
        if plane_normal.getTip()[2] == 1.:
            nodes = [list(p) for p in entity.points()]
        else:
            nodes = [list(entity.ocs().to_wcs(p)) for p in entity.points()]

        if entity.is_closed:
            nodes += [nodes[0]]

        p = Polyline()
        p.setNodes(nodes=nodes)
        return p

    @staticmethod
    def handleLine(entity: ezdxf.entities.Line) -> Optional[Polyline]:
        n0 = np.array(entity.dxf.start)
        n1 = np.array(entity.dxf.end)

        p = Polyline()
        p.setNodes(nodes=[n0, n1])
        return p



import os
import tempfile
from typing import Optional, Literal

import matplotlib.pyplot as plt
import matplotlib.image as mpimg
import numpy as np

from geogeometry.graphics.Figure3d import Figure3d
from geogeometry.graphics.shared.PlotFigure import PlotFigure


class Figure2d(PlotFigure):

    def __init__(self):
        super().__init__()
        self.figure: Optional[plt.Figure] = None
        self.ax: Optional[plt.Axes] = None

    def getFigure(self) -> Optional[plt.Figure]:
        return self.figure

    def getAx(self) -> Optional[plt.Axes]:
        return self.ax

    def setAxesLabels(self, projection_plane: Literal['xy', 'yz', 'xz'] = 'xy'):

        limits = self.calculateLimits()

        if projection_plane == 'xy':
            self.ax.set_xlabel('East [m]', fontsize=8)
            self.ax.set_ylabel('North [m]', fontsize=8)
            self.ax.set_xlim(limits[0][0], limits[1][0])
            self.ax.set_ylim(limits[0][1], limits[1][1])
        elif projection_plane == 'yz':
            self.ax.set_xlabel('North [m]', fontsize=8)
            self.ax.set_ylabel('Elevation [m]', fontsize=8)
            self.ax.set_xlim(limits[0][1], limits[1][1])
            self.ax.set_ylim(limits[0][2], limits[1][2])
        elif projection_plane == 'xz':
            self.ax.set_xlabel('East [m]', fontsize=8)
            self.ax.set_ylabel('Elevation [m]', fontsize=8)
            self.ax.set_xlim(limits[0][0], limits[1][0])
            self.ax.set_ylim(limits[0][2], limits[1][2])

    def plot(self, filepath: Optional[str] = None,
             projection_mode: Literal['2d', '3d'] = '2d',
             projection_plane: Literal['xy', 'yz', 'xz'] = 'xy') -> None:

        self.figure = plt.figure()
        self.ax = self.figure.add_subplot()

        if projection_mode == '2d':
            for e in self.elements:
                e.addToMatplotlibAx(ax=self.ax)
            self.ax.set_aspect("equal")

        else:
            fig3d = Figure3d()
            for e in self.elements:
                e.addToFigure3d(figure_3d=fig3d)

            tmpfile = tempfile.NamedTemporaryFile(suffix=".png", delete=False)
            fig3d.plot(filepath=tmpfile.name, projection_plane=projection_plane, show_axes=False)

            img = mpimg.imread(tmpfile.name)
            limits = self.calculateLimits()

            if projection_plane == 'xy':
                extent = [limits[0][0], limits[1][0], limits[0][1], limits[1][1]]
            elif projection_plane == 'yz':
                extent = [limits[0][1], limits[1][1], limits[0][2], limits[1][2]]
            elif projection_plane == 'xz':
                extent = [limits[0][0], limits[1][0], limits[0][2], limits[1][2]]

            self.ax.imshow(img, extent=extent)

            tmpfile.close()
            if os.path.exists(tmpfile.name):
                os.remove(tmpfile.name)

            self.setAxesLabels(projection_plane=projection_plane)

        if filepath is not None:
            plt.savefig(filepath, bbox_inches="tight", dpi=110)
        else:

            plt.show()

        plt.close(self.figure)



from typing import Optional

import numpy as np


class BlockProperties(object):

    def __init__(self):
        self.corners: Optional[np.ndarray] = None

    def setCorners(self, corners: np.ndarray) -> None:
        self.corners = corners

    def getCorners(self) -> np.ndarray:
        return self.corners



# def readStlFile(self, **kwargs):
from typing import Optional

import numpy as np
import trimesh

from geogeometry.geometry.model.GeometryModel import GeometryModel
from geogeometry.geometry.triangulation.Triangulation import Triangulation
from geogeometry.geometry.model.GeometryModelsCollection import GeometryModelsCollection


class StlReader(object):

    @staticmethod
    def readFile(filepath: str) -> Optional[GeometryModelsCollection]:

        mesh = trimesh.load(filepath)
        t_name = filepath.split(".")[0]

        t = Triangulation(name=t_name)
        t.setNodes(nodes=np.asarray(mesh.vertices))
        t.setFaces(faces=np.asarray(mesh.faces))

        model = GeometryModel(name=t_name)
        model.addTriangulation(t)

        models_collection = GeometryModelsCollection()
        models_collection.addElement(model)

        return models_collection



import csv
from typing import Optional

from geogeometry.geometry.model.GeometryModel import GeometryModel
from geogeometry.geometry.model.GeometryModelsCollection import GeometryModelsCollection


class CsvReader(object):

    @staticmethod
    def readFile(filepath: str, xlabel: str, ylabel: str, zlabel: str) -> Optional[GeometryModelsCollection]:

        pts_name = filepath.split(".")[0]

        points = []

        with open(filepath, mode='r') as file:
            reader = csv.reader(file)
            header = next(reader)

            xindex, yindex, zindex = header.index(xlabel), header.index(ylabel), header.index(zlabel)

            for row in reader:
                pt = [row[xindex], row[yindex], row[zindex]]
                points += [[float(c) for c in pt]]

        model = GeometryModel(name=pts_name)
        model.addPoints(points=points)

        models_collection = GeometryModelsCollection()
        models_collection.addElement(model)

        return models_collection



from typing import Optional, TYPE_CHECKING

from geogeometry.geometry.model.collection_components.GeometryModelsCollectionPlotter import GeometryModelsCollectionPlotter
from geogeometry.geometry.shared.BaseObjectCollection import BaseObjectCollection
from geogeometry.geometry.polyline.PolylinesCollection import PolylinesCollection
from geogeometry.geometry.triangulation.TriangulationsCollection import TriangulationsCollection

if TYPE_CHECKING:
    from geogeometry.geometry.model.GeometryModel import GeometryModel


class GeometryModelsCollection(BaseObjectCollection['GeometryModelsCollection', 'GeometryModel']):

    def __init__(self, name: Optional[str] = None):
        super().__init__(name=name)

        # self.plotter: GeometryModelsCollectionPlotter = GeometryModelsCollectionPlotter(self)

    def __str__(self) -> str:
        txt = f"GeometryModelsCollection(name={self.getName()}, models={len(self)})"
        return txt

    def deleteEmptyModels(self) -> None:
        empty_models = [e for e in self if not len(e)]
        for m in empty_models:
            self.deleteElement(identifier=m.getName())

    def getAllTriangulations(self) -> TriangulationsCollection:
        all_triangulations = TriangulationsCollection()

        for m in self:
            for t in m.getTriangulations():
                all_triangulations.addElement(t)

        return all_triangulations

    def getAllPolylines(self) -> PolylinesCollection:
        all_polylines = PolylinesCollection()

        for m in self:
            for p in m.getPolylines():
                all_polylines.addElement(p)

        return all_polylines



import os
import unittest
from geogeometry import CadReader


class TestReadCsvPoints(unittest.TestCase):

    def setUp(self):
        self.base_dir = os.path.dirname(os.path.abspath(__file__))
        self.resource_dir = os.path.join(self.base_dir, 'resources', 'csv_files')

        self.extension = "csv"

        self.file_001_path = os.path.join(self.resource_dir, "file_001." + self.extension)

    def test_file_001(self):
        models_collection = CadReader.readFile(self.file_001_path,
                                               xlabel="LocX [m]", ylabel="LocY [m]", zlabel="LocZ [m]")
        self.assertEqual(len(models_collection), 1)

        n_points = len(models_collection[0].getPoints())
        self.assertEqual(n_points, 4336)


if __name__ == "__main__":
    unittest.main()



from typing import Literal, Optional, Union

import numpy as np
import pyvista as pv

from geogeometry.graphics.shared.PlotFigure import PlotFigure


class Figure3d(PlotFigure):

    def __init__(self, engine: Literal['pyvista', 'paraview'] = 'pyvista'):
        super().__init__()

        self.engine: Literal['pyvista', 'paraview'] = engine

        self.plotter: Optional[Union[pv.Plotter]] = None

    def getEngine(self) -> Literal['pyvista', 'paraview']:
        return self.engine

    def setParallelPlaneView(self, projection_plane: Literal['xy', 'yz', 'xz']) -> None:
        self.plotter.camera.parallel_projection = True

        limits = self.calculateLimits()
        mid_point = np.average(limits, axis=0)

        xmin, ymin, zmin = limits[0]
        xmax, ymax, zmax = limits[1]

        if projection_plane == 'xy':
            # Set camera position above the triangulation and looking at the XY plane
            self.plotter.camera_position = [
                (0.5 * (xmin + xmax), 0.5 * (ymin + ymax), zmax + 100),  # Safe Z above the scene
                mid_point,  # Focus point
                (0, 1, 0)  # Up direction
            ]

            self.plotter.camera.SetParallelScale((ymax - ymin) / 2)

            # Calculate the aspect ratio for the screenshot
            x_range = xmax - xmin
            y_range = ymax - ymin
            aspect_ratio = x_range / y_range

        elif projection_plane == 'yz':
            # Set camera position above the triangulation and looking at the XY plane
            self.plotter.camera_position = [
                (xmax + 100, 0.5 * (ymin + ymax), 0.5 * (zmin + zmax)),  # Safe Z above the scene
                mid_point,  # Focus point
                (0, 0, 1)  # Up direction
            ]

            self.plotter.camera.SetParallelScale((zmax - zmin) / 2)

            # Calculate the aspect ratio for the screenshot
            y_range = ymax - ymin
            z_range = zmax - zmin
            aspect_ratio = y_range / z_range

        elif projection_plane == 'xz':
            # Set camera position above the triangulation and looking at the XY plane
            self.plotter.camera_position = [
                (0.5 * (xmin + xmax), ymax - 100, 0.5 * (zmin + zmax)),  # Safe Z above the scene
                mid_point,  # Focus point
                (0, 0, 1)  # Up direction
            ]

            self.plotter.camera.SetParallelScale((zmax - zmin) / 2)

            # Calculate the aspect ratio for the screenshot
            x_range = xmax - xmin
            z_range = zmax - zmin
            aspect_ratio = x_range / z_range

        # Adjust plotter window size for the correct aspect ratio
        height = 500  # Set height of the image (adjust as needed)
        width = int(height * aspect_ratio)
        self.plotter.window_size = [width, height]

    def plot(self, filepath: Optional[str] = None,
             projection_plane: Optional[Literal['xy', 'yz', 'xz']] = None,
             show_axes: bool = True) -> None:

        if self.plotter is None:
            if self.engine == 'pyvista':
                self.plotter = pv.Plotter(off_screen=filepath is not None)
            else:
                raise ValueError("Not yet implemented.")

        for e in self.elements:
            e.addToPyvistaPlotter(self.plotter)

        if projection_plane is not None:
            self.setParallelPlaneView(projection_plane=projection_plane)

        if show_axes:
            self.plotter.add_axes()

        if filepath is not None:
            self.plotter.screenshot(filepath)  # Save the plot to an image file
        else:
            self.plotter.show()




from typing import TYPE_CHECKING

import pyvista as pv

from geogeometry.geometry.shared.shared_components.BasePlotter import BasePlotter

if TYPE_CHECKING:
    import matplotlib.pyplot as plt
    from geogeometry.geometry.triangulation.TriangulationsCollection import TriangulationsCollection


class TriangulationsCollectionPlotter(BasePlotter):

    def __init__(self, triangulations_collection: 'TriangulationsCollection'):
        super().__init__(element=triangulations_collection)
        self.triangulations_collection: 'TriangulationsCollection' = triangulations_collection

    def addToMatplotlibAx(self, ax: 'plt.Axes') -> None:
        for t in self.triangulations_collection:
            t.addToMatplotlibAx(ax=ax)

    def addToPyvistaPlotter(self, plotter: 'pv.Plotter', **kwargs) -> None:
        for t in self.triangulations_collection:
            t.addToPyvistaPlotter(plotter)

    def addToParaviewPlotter(self, plotter, **kwargs) -> None: ...


from typing import Union, List, Optional, TYPE_CHECKING

import numpy as np

from geogeometry.geometry.model.components.GeometryModelMetrics import GeometryModelMetrics
from geogeometry.geometry.model.components.GeometryModelPlotter import GeometryModelPlotter
from geogeometry.geometry.shared.BaseObject import BaseObject
from geogeometry.geometry.model.components.GeometryModelProperties import GeometryModelProperties
from geogeometry.geometry.shared.observer.Observer import Observer

if TYPE_CHECKING:
    from geogeometry.geometry.shared.observer.Observable import Observable


class GeometryModel(BaseObject['GeometryModel'], GeometryModelProperties, Observer):

    def __init__(self, name: Optional[str] = None):
        super().__init__(name=name)

        self.getTriangulations().addObserver(observer=self)
        self.getPolylines().addObserver(observer=self)
        self.getPoints().addObserver(observer=self)

        self.metrics: GeometryModelMetrics = GeometryModelMetrics(self)
        self.plotter: GeometryModelPlotter = GeometryModelPlotter(self)

    def __len__(self) -> int:
        total = 0
        total += len(self.getTriangulations())
        total += len(self.getPolylines())
        return total

    def __str__(self) -> str:
        txt = (f"GeometryModel(name={self.getName()}, "
               f"triangulations={len(self.getTriangulations())}, "
               f"polylines={len(self.getPolylines())}, "
               f"points={len(self.getPoints())})")
        return txt

    # def calculateMetrics(self):
    #     self.calculateLimits()
    #     self.calculateCentroid()
    #
    # def calculateLimits(self):
    #     all_limits = [lim for t in self.getTriangulations() for lim in t.getLimits()]
    #     all_limits += [lim for p in self.getPolylines() for lim in p.getLimits()]
    #
    #     _min, _max = np.min(all_limits, axis=0), np.max(all_limits, axis=0)
    #     self.setLimits(limits=np.array([_min, _max]))
    #
    # def calculateCentroid(self) -> None:
    #     all_centers = [t.getCentroid() for t in self.getTriangulations()]
    #     all_centers += [p.getCentroid() for p in self.getPolylines()]
    #
    #     center = np.average(all_centers, axis=0)
    #     self.setCentroid(centroid=center)

    def onElementChange(self, observable: 'Observable'):
        self.calculateMetrics()
        self.notifyObservers()



from typing import TYPE_CHECKING

from geogeometry.geometry.shared.shared_components.BasePlotter import BasePlotter

if TYPE_CHECKING:
    import matplotlib.pyplot as plt
    import pyvista as pv
    from geogeometry.geometry.vector.Vector import Vector


class VectorPlotter(BasePlotter):

    def __init__(self, vector: 'Vector'):
        super().__init__(element=vector)
        self.vector: 'Vector' = vector

    def addToMatplotlibAx(self, ax: 'plt.Axes') -> None: ...

    def addToPyvistaPlotter(self, plotter: 'pv.Plotter', **kwargs) -> None: ...

    def addToParaviewPlotter(self, plotter, **kwargs) -> None: ...



from typing import TypeVar, TYPE_CHECKING

from geogeometry.geometry.shared.shared_components.BasePlotter import BasePlotter

if TYPE_CHECKING:
    import pyvista as pv
    import matplotlib.pyplot as plt
    from geogeometry.geometry.shared.BaseObjectCollection import BaseObjectCollection


CollectionType = TypeVar('CollectionType', bound='BaseObjectCollection')


class BaseCollectionPlotter(BasePlotter):

    def __init__(self, collection: CollectionType):
        super().__init__(element=collection)
        self.collection: CollectionType = collection

    def addToMatplotlibAx(self, ax: 'plt.Axes') -> None:
        for p in self.collection:
            p.addToMatplotlibAx(ax)

    def addToPyvistaPlotter(self, plotter: 'pv.Plotter', **kwargs) -> None:
        for p in self.collection:
            p.addToPyvistaPlotter(plotter)

    def addToParaviewPlotter(self, plotter, **kwargs) -> None: ...



from typing import Optional, TYPE_CHECKING

import numpy as np

from geogeometry.geometry.shared.explicit_geometry.ExplicitGeometryMetrics import ExplicitGeometryMetrics
from geogeometry.geometry.shared.shared_components.BaseMetrics import BaseMetrics

if TYPE_CHECKING:
    from geogeometry.geometry.polyline.Polyline import Polyline


class PolylineMetrics(BaseMetrics):

    def __init__(self, polyline: 'Polyline'):
        self.polyline: 'Polyline' = polyline

        self.length: Optional[float] = None
        self.closed: Optional[bool] = None
        self.mid_point: Optional[np.ndarray] = None

    def getLength(self) -> Optional[float]:
        return self.length

    def isClosed(self) -> Optional[bool]:
        return self.closed

    def getMidPoints(self) -> Optional[np.ndarray]:
        return self.mid_point

    def calculateMetrics(self) -> None:
        self.calculateClosed()
        self.calculateLength()
        self.calculateMidPoint()

        # if self.polyline.getDimensions() == 2 and self.polyline.isClosed():
        #     self.calculate2DArea()

    def calculateClosed(self) -> None:
        nodes = self.polyline.getNodes()
        self.closed = False
        if not np.linalg.norm(nodes[0] - nodes[-1]):
            self.closed = True

    def calculateLength(self) -> None:
        nodes = self.polyline.getNodes()
        ds = [np.linalg.norm(n - nodes[i + 1]) for i, n in enumerate(nodes[:-1])]
        self.length = np.sum(ds)

    def calculateMidPoint(self) -> None:
        """
        Mid at (total length) / 2
        """
        self.mid_point = self.polyline.getPointAtDistanceFromOrigin(distance=self.polyline.getLength() / 2.)

    # def calculate2DArea(self) -> None:
    #     """
    #     Considers XY space, even if nodes dimensions is greater than 2.
    #     Gets area even if the polyline is open.
    #     """
    #     nodes = self.polyline.getNodes()
    #     if self.polyline.isClosed():
    #         x, y = nodes[:-1][:, 0], nodes[:-1][:, 1]
    #     else:
    #         x, y = nodes[:, 0], nodes[:, 1]
    #
    #     area = 0.5 * np.abs(np.dot(x, np.roll(y, 1)) - np.dot(y, np.roll(x, 1)))
    #
    #     self.polyline.setArea(area=area)





from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from geogeometry.geometry.shared.observer.Observable import Observable


class Observer(ABC):

    @abstractmethod
    def onElementChange(self, observable: 'Observable'): ...




from typing import TYPE_CHECKING, Optional

import pyvista as pv


from geogeometry.geometry.shared.shared_components.BasePlotter import BasePlotter

if TYPE_CHECKING:
    import matplotlib.pyplot as plt
    from geogeometry.geometry.polyline.PolylinesCollection import PolylinesCollection


class PolylinesCollectionPlotter(BasePlotter):

    def __init__(self, polylines_collection: 'PolylinesCollection'):
        super().__init__(element=polylines_collection)
        self.polylines_collection: 'PolylinesCollection' = polylines_collection

    def addToMatplotlibAx(self, ax: 'plt.Axes') -> None:
        for p in self.polylines_collection:
            p.addToMatplotlibAx(ax)

    def addToPyvistaPlotter(self, plotter: 'pv.Plotter', **kwargs) -> None:
        for p in self.polylines_collection:
            p.addToPyvistaPlotter(plotter)

    def addToParaviewPlotter(self, plotter, **kwargs) -> None: ...



from typing import Optional, TYPE_CHECKING, Union, List

import numpy as np

from geogeometry.geometry.points.PointsCollection import PointsCollection
from geogeometry.geometry.polyline.PolylinesCollection import PolylinesCollection
from geogeometry.geometry.triangulation.TriangulationsCollection import TriangulationsCollection

if TYPE_CHECKING:
    from geogeometry.geometry.polyline.Polyline import Polyline
    from geogeometry.geometry.triangulation.Triangulation import Triangulation
    from geogeometry.geometry.model.components.GeometryModelMetrics import GeometryModelMetrics


class GeometryModelProperties(object):

    def __init__(self):
        super().__init__()
        self.triangulations: TriangulationsCollection = TriangulationsCollection()
        self.polylines: PolylinesCollection = PolylinesCollection()
        self.points: PointsCollection = PointsCollection()

        self.metrics: Optional['GeometryModelMetrics'] = None

    def getTriangulations(self) -> Optional[TriangulationsCollection]:
        return self.triangulations

    def getPolylines(self) -> Optional[PolylinesCollection]:
        return self.polylines

    def getPoints(self) -> Optional[PointsCollection]:
        return self.points

    def getLimits(self) -> Optional[np.ndarray]:
        return self.metrics.getLimits()

    def getCenter(self) -> Optional[np.ndarray]:
        return self.metrics.getCenter()

    def addTriangulation(self, triangulation: 'Triangulation') -> None:
        self.triangulations.addElement(triangulation)

    def addPolyline(self, polyline: 'Polyline') -> None:
        self.polylines.addElement(polyline)

    def addPoints(self, points: Union[List, np.ndarray]) -> None:
        self.points.setPoints(points)



import os
import unittest
from geogeometry import CadReader


class TestReadRhinoMixed(unittest.TestCase):

    def setUp(self):
        self.base_dir = os.path.dirname(os.path.abspath(__file__))
        self.resource_dir = os.path.join(self.base_dir, 'resources', 'rhino_files')

        self.extension = "3dm"

        self.file_006_path = os.path.join(self.resource_dir, "file_006." + self.extension)

    def test_file_006(self):
        models_collection = CadReader.readFile(self.file_006_path)
        self.assertEqual(len(models_collection), 2)

        n_triangulations = len(models_collection.getAllTriangulations())
        self.assertEqual(n_triangulations, 2)

        n_polylines = len(models_collection.getAllPolylines())
        self.assertEqual(n_polylines, 3)



from typing import Optional, Union, List, TYPE_CHECKING

import numpy as np

from geogeometry.geometry.circle.components.CircleMetrics import CircleMetrics
from geogeometry.geometry.circle.components.CirclePlotter import CirclePlotter
from geogeometry.geometry.circle.components.CircleProperties import CircleProperties
from geogeometry.geometry.shared.BaseObject import BaseObject

if TYPE_CHECKING:
    from geogeometry.geometry.plane.Plane import Plane


class Circle(BaseObject, CircleProperties):

    def __init__(self,
                 center: Union[List, np.ndarray],
                 radius: float,
                 plane: Optional['Plane'] = None,
                 name: Optional[str] = None):

        super().__init__(name=name)

        self.metrics: CircleMetrics = CircleMetrics(self)
        self.plotter: CirclePlotter = CirclePlotter(self)

        self.setCenter(center=center)
        self.setRadius(radius=radius)
        self.setPlane(plane=plane)



import math
from typing import Callable, TYPE_CHECKING, Tuple, Union, List, Literal

import numpy as np

from geogeometry.geometry.vector.Vector import Vector
from geogeometry.geometry.operations.Rotations import Rotations

if TYPE_CHECKING:
    from geogeometry.geometry.plane.Plane import Plane


PointType = Union[List, np.ndarray]
PlaneInputsType = Tuple[np.ndarray, Vector]


class PlaneFactory(object):

    @staticmethod
    def createFromThreePoints(n0: PointType, n1: PointType, n2: PointType) -> PlaneInputsType:
        if isinstance(n0, list):
            n0 = np.array(n0)
        if isinstance(n1, list):
            n1 = np.array(n1)
        if isinstance(n2, list):
            n2 = np.array(n2)

        vec01 = n1 - n0  # Vector from origin to node 1.
        vec02 = n2 - n0  # Vector from origin to node 2.

        # noinspection PyUnreachableCode
        normal_vector = Vector(tip=np.cross(vec01, vec02))

        return n0, normal_vector

    @staticmethod
    def createFromDipAndDipdir(origin: PointType, dip: float, dipdir: float) -> PlaneInputsType:
        if isinstance(origin, list):
            origin = np.array(origin)

        if dip < 0:
            raise ValueError("Dip value must be between 0 and 90°.")

        normal_vector = Vector(tip=[0, 0, 1])

        Rx = Rotations.calculateXRotationMatrix(angle=-math.radians(dip))
        Rz = Rotations.calculateZRotationMatrix(angle=-math.radians(dipdir))

        normal_vector.rotateByRotationMatrix(rotation_matrix=Rx)
        normal_vector.rotateByRotationMatrix(rotation_matrix=Rz)

        return origin, normal_vector

    @staticmethod
    def createAsOrthogonal(axis: Literal['x', 'y', 'z'], level: float) -> PlaneInputsType:
        if axis == 'x':
            tip = [1, 0, 0]
        elif axis == 'y':
            tip = [0, 1, 0]
        elif axis == 'z':
            tip = [0, 0, 1]
        else:
            raise ValueError(f"Wrong axis definition: '{axis}'")

        tip = np.array(tip)
        normal_vector = Vector(tip=tip)
        origin = tip * level

        return origin, normal_vector

    @staticmethod
    def createAsVerticalProjection(n0: PointType, n1: PointType) -> PlaneInputsType:
        return PlaneFactory.createFromThreePoints(n0=list(n0)+[0], n1=list(n1)+[0], n2=list(n1)+[1])




from typing import Optional, Union, List

import numpy as np

from geogeometry.geometry.block.components.BlockProperties import BlockProperties
from geogeometry.geometry.shared.explicit_geometry.ExplicitGeometry import ExplicitGeometry


class Block(ExplicitGeometry, BlockProperties):

    def __init__(self,
                 corners: Union[List, np.ndarray],
                 name: Optional[str] = None):
        super().__init__(name=name)

        corners = np.array([np.min(corners, axis=0), np.max(corners, axis=0)])

        self.setCorners(corners=corners)
        self.setNodes(nodes=corners)




from typing import Optional

import numpy as np

from geogeometry.geometry.shared.BaseObject import BaseObject


class Sphere(BaseObject, SphereProperties):

    def __init__(self,
                 center: np.ndarray,
                 radius: float,
                 name: Optional[str] = None):
        super().__init__(name=name)

        self.setCenter(center=center)
        self.setRadius(radius=radius)


from typing import Union, List, Optional, TYPE_CHECKING, Iterable, Literal

import numpy as np

from geogeometry.geometry.polyline.components.PolylineTransformer import PolylineTransformer

from geogeometry.geometry.polyline.components.PolylinePlotter import PolylinePlotter
from geogeometry.geometry.polyline.components.PolylineProperties import PolylineProperties

from geogeometry.geometry.polyline.components.PolylineMetrics import PolylineMetrics
from geogeometry.geometry.polyline.components.PolylineQuerier import PolylineQuerier
from geogeometry.geometry.shared.explicit_geometry.ExplicitGeometry import ExplicitGeometry


class Polyline(ExplicitGeometry, PolylineProperties):

    def __init__(self, name: Optional[str] = None):
        super().__init__(name=name)

        self.metrics: PolylineMetrics = PolylineMetrics(self)
        self.querier: PolylineQuerier = PolylineQuerier(self)
        self.transformer: PolylineTransformer = PolylineTransformer(self)

        self.plotter: PolylinePlotter = PolylinePlotter(self)

    def __len__(self) -> int:
        if self.getNodes() is None:
            return 0
        return len(self.getNodes())

    def __iter__(self) -> Iterable[np.ndarray]:
        for n in self.getNodes():
            yield n

    def __str__(self) -> str:
        txt = f"Polyline(name={self.getName()}, nodes={len(self)}, closed={self.isClosed()})"
        return txt

    def setNodes(self, nodes: Union[List, np.ndarray]) -> None:
        super().setNodes(nodes=nodes)
        self.segments = [[j, j + 1] for j, n in enumerate(self.nodes[:-1])]

        self.metrics.calculateMetrics()

    # QUERIES
    def getPointAtDistanceFromOrigin(self, distance: float) -> Optional[np.ndarray]:
        return self.querier.getPointAtDistanceFromOrigin(distance=distance)



import unittest
from geogeometry import Triangulation, Vector, Rotations, Plane


class TestBaseGeometry(unittest.TestCase):

    def setUp(self):
        pass

    def test_triangulation_creation(self):
        t = Triangulation()
        self.assertIsNotNone(t)
        self.assertIsInstance(t, Triangulation)

    def test_vectors_rotations(self):
        v1 = Vector(tip=[0., 0., 1.])
        v2 = Vector(tip=[0, 1, 0])

        R = Rotations.calculateRotationMatrixFromVectors(v0=v1, v1=v2)

        v1.rotateByRotationMatrix(R)

        self.assertEqual(v1, v2)

    def test_planes_creation(self):
        test_vector = Vector(tip=[1, 0, 0])

        p = Plane.createFromThreePoints(n0=[0, 0, 0], n1=[0, 1, 0], n2=[0, 0, 1])
        self.assertEqual(p.getNormalVector(), test_vector)

        p = Plane.createFromDipAndDipdir(origin=[0, 0, 0], dip=90., dipdir=90.)
        self.assertEqual(p.getNormalVector(), test_vector)

        p = Plane.createAsOrthogonal(axis='x', level=500.)
        self.assertEqual(p.getNormalVector(), test_vector)

        p = Plane.createAsVerticalProjection(n0=[0, 0], n1=[0, 1])
        self.assertEqual(p.getNormalVector(), test_vector)


if __name__ == "__main__":
    unittest.main()



from typing import Union, List

import numpy as np

from geogeometry.geometry.vector.Vector import Vector


class Collinearity(object):

    @staticmethod
    def isNodeWithinSegment(point: np.ndarray, segment: Union[List[float], np.ndarray],
                            cross_tolerance: float = 1e-6, dot_tolerance: float = 1e-12) -> bool:
        if isinstance(segment, list):
            segment = np.array(segment)

        segment_vector = Vector(origin=segment[0], tip=segment[1])
        test_vector = Vector(origin=segment[0], tip=point)

        if test_vector.getLength() > segment_vector.getLength():
            return False

        # noinspection PyUnreachableCode
        cross_vector = Vector(tip=np.cross(segment_vector.getUnitVector(), test_vector.getUnitVector()))

        # Si colinean checkeo que node esta en segment
        if cross_vector.getLength() < cross_tolerance:
            # Debe ser mayor que 0 para que el punto este dentro del segment.
            dot = np.dot(segment_vector.getUnitVector(), test_vector.getUnitVector())
            if dot >= -dot_tolerance:  # - toleran ya que puede estar muy cerca por el otro lado
                return True

        return False



from typing import TYPE_CHECKING, Optional

import numpy as np
import pyvista as pv
import matplotlib.tri as tri

from geogeometry.geometry.shared.shared_components.BasePlotter import BasePlotter

if TYPE_CHECKING:
    import matplotlib.pyplot as plt
    from geogeometry.geometry.triangulation.Triangulation import Triangulation


class TriangulationPlotter(BasePlotter):

    def __init__(self, triangulation: 'Triangulation'):
        super().__init__(element=triangulation)
        self.triangulation: 'Triangulation' = triangulation

    def addToMatplotlibAx(self, ax: 'plt.Axes') -> None:
        nodes = self.triangulation.getNodes()
        faces = self.triangulation.getFaces()

        t = tri.Triangulation(nodes[:, 0], nodes[:, 1], faces)
        ax.triplot(t, color=self.triangulation.getColor())

    def _getPyVistaTriangulation(self) -> pv.PolyData:
        pv_faces = np.hstack((np.full((self.triangulation.getFaces().shape[0], 1), 3),
                              self.triangulation.getFaces()))

        return pv.PolyData(self.triangulation.getNodes(), faces=pv_faces)

    def addToPyvistaPlotter(self, plotter: 'pv.Plotter', **kwargs) -> None:
        plotter.add_mesh(self._getPyVistaTriangulation(),
                         show_edges=False,
                         color=self.triangulation.getColor(),
                         opacity=self.triangulation.getOpacity())

    def addToParaviewPlotter(self, plotter, **kwargs) -> None: ...



from typing import Optional, Union, List, Tuple, Callable, Literal

import numpy as np

from geogeometry.geometry.plane.components.PlaneFactory import PlaneFactory
from geogeometry.geometry.plane.components.PlaneProperties import PlaneProperties
from geogeometry.geometry.shared.BaseObject import BaseObject
from geogeometry.geometry.vector.Vector import Vector

PointType = Union[List, np.ndarray]
PlaneInputsType = Tuple[np.ndarray, Vector]


def createFromInputs(func: Callable[..., PlaneInputsType]) -> Callable[..., 'Plane']:
    def inner(*args, **kwargs) -> 'Plane':
        origin, normal_vector = func(*args, **kwargs)
        p = Plane(origin=origin, normal_vector=normal_vector)
        return p

    return inner


class Plane(BaseObject['Plane'], PlaneProperties):

    def __init__(self,
                 origin: PointType,
                 normal_vector: 'Vector',
                 name: Optional[str] = None):
        super().__init__(name=name)

        self.setOrigin(origin=origin)
        self.setNormalVector(normal_vector=normal_vector)

    def __str__(self) -> str:
        txt = f"Plane(name={self.getName()}, origin={self.getOrigin()}, normal_vector={self.getNormalVector()})"
        return txt

    @staticmethod
    @createFromInputs
    def createFromThreePoints(n0: PointType, n1: PointType, n2: PointType) -> PlaneInputsType:
        return PlaneFactory.createFromThreePoints(n0=n0, n1=n1, n2=n2)

    @staticmethod
    @createFromInputs
    def createFromDipAndDipdir(origin: PointType, dip: float, dipdir: float) -> PlaneInputsType:
        return PlaneFactory.createFromDipAndDipdir(origin=origin, dip=dip, dipdir=dipdir)

    @staticmethod
    @createFromInputs
    def createAsOrthogonal(axis: Literal['x', 'y', 'z'], level: float) -> PlaneInputsType:
        return PlaneFactory.createAsOrthogonal(axis=axis, level=level)

    @staticmethod
    @createFromInputs
    def createAsVerticalProjection(n0: PointType, n1: PointType) -> PlaneInputsType:
        return PlaneFactory.createAsVerticalProjection(n0=n0, n1=n1)



from typing import TypeVar, TYPE_CHECKING, Optional

import numpy as np

from geogeometry.geometry.shared.shared_components.BaseMetrics import BaseMetrics

if TYPE_CHECKING:
    from geogeometry.geometry.shared.BaseObjectCollection import BaseObjectCollection


CollectionType = TypeVar('CollectionType', bound='BaseObjectCollection')


class BaseCollectionMetrics(BaseMetrics):

    def __init__(self, collection: CollectionType):
        self.collection: CollectionType = collection

        self.limits: Optional[np.ndarray] = None
        self.center: Optional[np.ndarray] = None

    def getLimits(self) -> Optional[np.ndarray]:
        return self.limits

    def getCenter(self) -> Optional[np.ndarray]:
        return self.center

    def calculateMetrics(self) -> None:
        for e in self.collection:
            if e.getLimits() is not None:
                break
        else:
            return

        self.calculateLimits()
        self.calculateCenter()

    def calculateLimits(self) -> None:
        all_limits = [lim for e in self.collection if e.getLimits() is not None for lim in e.getLimits()]
        _min, _max = np.min(all_limits, axis=0), np.max(all_limits, axis=0)
        self.limits = np.array([_min, _max])

    def calculateCenter(self) -> None:
        all_centers = [e.getCenter() for e in self.collection if e.getCenter() is not None]
        self.center = np.average(all_centers, axis=0)



from typing import TYPE_CHECKING, Optional

import numpy as np
import pyvista as pv
import matplotlib.pyplot as plt

from geogeometry.geometry.shared.shared_components.BasePlotter import BasePlotter

if TYPE_CHECKING:
    from geogeometry.geometry.polyline.Polyline import Polyline


class PolylinePlotter(BasePlotter):

    def __init__(self, polyline: 'Polyline'):
        super().__init__(element=polyline)
        self.polyline: 'Polyline' = polyline

    def addToMatplotlibAx(self, ax: 'plt.Axes') -> None:
        nodes = self.polyline.getNodes()
        ax.plot(nodes[:, 0], nodes[:, 1])

    def _getPyVistaPolyline(self) -> pv.PolyData:
        lines = np.hstack([[len(self.polyline.getNodes())], np.arange(len(self.polyline.getNodes()))])
        return pv.PolyData(self.polyline.getNodes(), lines=lines)

    def addToPyvistaPlotter(self, plotter: 'pv.Plotter', **kwargs) -> None:
        plotter.add_mesh(self._getPyVistaPolyline(),
                         color=self.polyline.getColor(),
                         opacity=self.polyline.getOpacity(),
                         line_width=self.polyline.getLineWidth())

    def addToParaviewPlotter(self, plotter, **kwargs) -> None: ...



from typing import Optional, TYPE_CHECKING, Literal

from geogeometry.geometry.shared.BaseObjectCollection import BaseObjectCollection
from geogeometry.geometry.triangulation.collection_components.TriangulationsCollectionPlotter import \
    TriangulationsCollectionPlotter

if TYPE_CHECKING:
    from geogeometry.geometry.triangulation.Triangulation import Triangulation


class TriangulationsCollection(BaseObjectCollection['TriangulationCollection', 'Triangulation']):

    def __init__(self, name: Optional[str] = None):
        super().__init__(name=name)

        # self.plotter: TriangulationsCollectionPlotter = TriangulationsCollectionPlotter(self)

    def __str__(self) -> str:
        txt = f"TriangulationsCollection(name={self.getName()}, triangulations={len(self)})"
        return txt



from setuptools import setup, find_packages

# TESTING
# python -m unittest discover -s tests
# PRE-COMMIT TESTING (with .yaml file ready)
# pre-commit install

# BUILD
# python setup.py sdist bdist_wheel
# PUBLISH
# twine upload dist/*

setup(
    name="geogeometry",
    version="0.0.4",
    description="A Python geometrical library made for geotechnical engineering.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Jorge Martinez",
    author_email="jmartinez@gmintec.com",
    license="MIT",
    packages=find_packages(),
    install_requires=open("requirements.txt").read().splitlines(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.10",
)



from typing import Optional, Callable, List, Union, TYPE_CHECKING, Literal

import numpy as np

from geogeometry.geometry.shared.explicit_geometry.ExplicitGeometry import ExplicitGeometry

from geogeometry.geometry.vector.components.VectorMetrics import VectorMetrics
from geogeometry.geometry.vector.components.VectorPlotter import VectorPlotter
from geogeometry.geometry.vector.components.VectorProperties import VectorProperties
from geogeometry.geometry.vector.components.VectorQuerier import VectorQuerier
from geogeometry.geometry.vector.components.VectorTransformer import VectorTransformer


def nodesToVector(func: Callable[..., List[np.ndarray]]) -> Callable[..., 'Vector']:
    def inner(vector: 'Vector', *args, **kwargs) -> 'Vector':
        nodes = func(vector, *args, **kwargs)
        return Vector(origin=nodes[0], tip=nodes[1])

    return inner


class Vector(ExplicitGeometry, VectorProperties):

    def __init__(self,
                 tip: Union[List, np.ndarray],
                 origin: Optional[Union[List, np.ndarray]] = None,
                 name: Optional[str] = None):

        super().__init__(name=name)

        self.metrics: VectorMetrics = VectorMetrics(self)
        self.querier: VectorQuerier = VectorQuerier(self)
        self.transformer: VectorTransformer = VectorTransformer(self)
        self.plotter: VectorPlotter = VectorPlotter(self)

        if origin is None:
            origin = np.zeros(len(tip))

        self.setNodes(nodes=[origin, tip])

    def __eq__(self, other_vector: 'Vector') -> bool:
        diff = self.getNodes() - other_vector.getNodes()
        if np.linalg.norm(diff):
            return False
        return True

    def __str__(self) -> str:
        txt = f"Vector(origin={self.getOrigin()}, tip={self.getTip()})"
        return txt

    def setNodes(self, nodes: Union[List, np.ndarray]) -> None:
        super().setNodes(nodes=nodes)

        if len(nodes) != 2:
            raise ValueError("Trying to define an explicit vector without 2 nodes.")

        self.setOrigin(origin=self.getNodes()[0])
        self.setTip(tip=self.getNodes()[1])

        self.metrics.calculateMetrics()

    # QUERIER
    def calculateAngleWithVector(self, other_vector: 'Vector') -> float:
        return self.querier.calculateAngleWithVector(other_vector=other_vector)

    def isVectorParallel(self, other_vector: 'Vector') -> bool:
        diff = self.getUnitVector() - other_vector.getUnitVector()
        if np.linalg.norm(diff):
            return False
        return True

    # TRANSFORMATIONS
    def reverse(self) -> None:
        self.transformer.reverse()



import os
import tempfile
import unittest
from geogeometry import CadReader


class TestPlot3d(unittest.TestCase):

    def setUp(self):
        self.base_dir = os.path.dirname(os.path.abspath(__file__))
        self.resource_dir = os.path.join(self.base_dir, '..', 'test_io', 'resources', 'dxf_files')

        self.file_001_path = os.path.join(self.resource_dir, "file_001.dxf")  # Triangulations
        self.file_004_path = os.path.join(self.resource_dir, "file_004.dxf")  # Polylines
        self.file_006_path = os.path.join(self.resource_dir, "file_006.dxf")  # Mixed

    def test_file_001(self):
        models_collection = CadReader.readFile(self.file_001_path)

        t = models_collection[0].getTriangulations()[0]

        tmpfile = tempfile.NamedTemporaryFile(suffix=".png", delete=False)
        t.plot3d(screenshot=True, filepath=tmpfile.name)

        self.assertTrue(os.path.exists(tmpfile.name))

        tmpfile.close()
        if os.path.exists(tmpfile.name):
            os.remove(tmpfile.name)

    def test_file_004(self):
        models_collection = CadReader.readFile(self.file_004_path)

        drillholes = models_collection.getAllPolylines()

        tmpfile = tempfile.NamedTemporaryFile(suffix=".png", delete=False)
        drillholes.plot3d(filepath=tmpfile.name)

        self.assertTrue(os.path.exists(tmpfile.name))

        tmpfile.close()
        if os.path.exists(tmpfile.name):
            os.remove(tmpfile.name)

    def test_file_006(self):
        models_collection = CadReader.readFile(self.file_006_path)

        tmpfile = tempfile.NamedTemporaryFile(suffix=".png", delete=False)
        models_collection.plot3d(filepath=tmpfile.name)

        self.assertTrue(os.path.exists(tmpfile.name))

        tmpfile.close()
        if os.path.exists(tmpfile.name):
            os.remove(tmpfile.name)


if __name__ == "__main__":
    unittest.main()



from typing import Optional, TYPE_CHECKING

import numpy as np

from geogeometry.geometry.cylinder.components.CylinderProperties import CylinderProperties
from geogeometry.geometry.shared.BaseObject import BaseObject

if TYPE_CHECKING:
    from geogeometry import Vector


class Cylinder(BaseObject, CylinderProperties):

    def __init__(self,
                 radius: float,
                 origin: np.ndarray,
                 axis_vector: 'Vector',
                 length: float,
                 name: Optional[str] = None):
        super().__init__(name=name)

        self.setRadius(radius=radius)
        self.setOrigin(origin=origin)
        self.setAxisVector(axis_vector=axis_vector)
        self.setLength(length=length)



from typing import Optional, Union, List, Literal, TYPE_CHECKING

import numpy as np

from geogeometry.geometry.polyline.components.PolylinePlotProperties import PolylinePlotProperties

if TYPE_CHECKING:
    from geogeometry.geometry.polyline.components.PolylineMetrics import PolylineMetrics


class PolylineProperties(PolylinePlotProperties):

    def __init__(self):
        super().__init__()

        # self.nodes: Optional[np.ndarray] = None
        self.segments: Optional[np.ndarray] = None

        self.metrics: Optional['PolylineMetrics'] = None

        # self.closed: Optional[bool] = None
        # self.length: Optional[float] = None
        #
        # self.mid_point: Optional[np.ndarray] = None
        #
        # # 2D properties
        # self.area: Optional[float] = None

        # Plot properties
        # self.line_width: float = 2.

    # def setNodes(self, nodes: Union[List, np.ndarray]) -> None:
    #     self.nodes = np.array(nodes)
    #     self.segments = [[j, j + 1] for j, n in enumerate(self.nodes[:-1])]

    # def setClosed(self, closed: bool) -> None:
    #     self.closed = closed
    #
    # def setLength(self, length: float) -> None:
    #     self.length = length
    #
    # def setMidPoint(self, mid_point: np.ndarray) -> None:
    #     self.mid_point = mid_point
    #
    # def setArea(self, area: float) -> None:
    #     self.area = area

    # def setLineWidth(self, line_width: float) -> None:
    #     self.line_width = line_width

    # def getNodes(self) -> Optional[np.ndarray]:
    #     return self.nodes

    def getSegments(self) -> Optional[np.ndarray]:
        return self.segments

    def isClosed(self) -> Optional[bool]:
        return self.metrics.isClosed()

    def getLength(self) -> Optional[float]:
        return self.metrics.getLength()

    def getMidPoints(self) -> Optional[np.ndarray]:
        return self.mid_point

    def getArea(self) -> Optional[float]:
        return self.area

    # def getLineWidth(self) -> float:
    #     return self.line_width



from typing import TYPE_CHECKING

from geogeometry.geometry.shared.shared_components.BasePlotter import BasePlotter

from matplotlib.patches import Circle

if TYPE_CHECKING:
    import matplotlib.pyplot as plt
    import pyvista as pv
    from geogeometry.geometry.circle.Circle import Circle


class CirclePlotter(BasePlotter):

    def __init__(self, circle: 'Circle'):
        super().__init__(element=circle)
        self.circle: 'Circle' = circle

    def addToMatplotlibAx(self, ax: 'plt.Axes') -> None:
        circle = Circle(self.circle.getCenter(),
                        self.circle.getRadius(),
                        edgecolor='blue',
                        facecolor='none',
                        linewidth=2)
        ax.add_patch(circle)

    def addToPyvistaPlotter(self, plotter: 'pv.Plotter', **kwargs) -> None: ...

    def addToParaviewPlotter(self, plotter, **kwargs) -> None: ...



from typing import TYPE_CHECKING, Optional

from geogeometry.geometry.operations.Angles import Angles

if TYPE_CHECKING:
    from geogeometry.geometry.vector.Vector import Vector


class VectorQuerier(object):

    def __init__(self, vector: 'Vector'):
        self.vector = vector

    def calculateAngleWithVector(self, other_vector: 'Vector') -> float:
        return Angles.calculateAngleFromThreePoints(n0=self.vector.getUnitVector(),
                                                    n1=[0, 0, 0],
                                                    n2=other_vector.getUnitVector())



from typing import Optional, Union, List, TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from geogeometry.geometry.plane.Plane import Plane


class CircleProperties(object):

    def __init__(self):
        self.center: Optional[np.ndarray] = None
        self.radius: Optional[float] = None
        self.plane: Optional['Plane'] = None

    def setCenter(self, center: Union[List, np.ndarray]) -> None:
        if isinstance(center, list):
            center = np.array(center)
        self.center = center

    def setRadius(self, radius: float) -> None:
        self.radius = radius

    def setPlane(self, plane: 'Plane') -> None:
        self.plane = plane

    def getCenter(self) -> np.ndarray:
        return self.center

    def getRadius(self) -> float:
        return self.radius

    def getPlane(self) -> Optional['Plane']:
        return self.plane



class BasePlotProperties(object):

    def __init__(self):
        super().__init__()

        self.color: str = 'red'
        self.opacity: float = 1.0

    def setColor(self, color: str) -> None:
        self.color = color

    def setOpacity(self, opacity: float) -> None:
        self.opacity = opacity

    def getColor(self) -> str:
        return self.color

    def getOpacity(self) -> float:
        return self.opacity



from abc import ABC, abstractmethod
from typing import Union, TYPE_CHECKING

if TYPE_CHECKING:
    from geogeometry.geometry.shared.BaseObject import BaseObject
    from geogeometry.geometry.shared.BaseObjectCollection import BaseObjectCollection


class BaseMetrics(ABC):

    @abstractmethod
    def calculateMetrics(self) -> None: ...



from typing import TYPE_CHECKING

import pyvista as pv

from geogeometry.geometry.shared.shared_components.BasePlotter import BasePlotter

if TYPE_CHECKING:
    import matplotlib.pyplot as plt
    from geogeometry.geometry.points import PointsCollection


class PointsCollectionPlotter(BasePlotter):

    def __init__(self, points_collection: 'PointsCollection'):
        super().__init__(element=points_collection)
        self.points_collection: 'PointsCollection' = points_collection

    def addToMatplotlibAx(self, ax: 'plt.Axes') -> None:
        pts = self.points_collection.getPoints()
        ax.scatter(pts[:, 0], pts[:, 1])

    def _getPyVistaPointsCloud(self) -> pv.PolyData:
        return pv.PolyData(self.points_collection.getPoints())

    def addToPyvistaPlotter(self, plotter: 'pv.Plotter', **kwargs) -> None:
        plotter.add_mesh(self._getPyVistaPointsCloud(),
                         color=self.points_collection.getColor(),
                         opacity=self.points_collection.getOpacity(),
                         point_size=10,
                         render_points_as_spheres=True)

    def addToParaviewPlotter(self, plotter, **kwargs) -> None: ...
